from .core import Primitive, synchronized, PyThread
from .dequeue import DEQueue
from .task_queue import Task, TaskQueue
from .promise import Promise
import sys, traceback

class _WorkerTask(Task):
    
    def __init__(self, processor, item, promise):
        Task.__init__(self)
        self.Processor = processor
        self.Item = item
        self.Promise = promise
        
    def run(self):
        out = self.Processor._process(self.Item, self.Promise)
        
class _CloseOutputThread(Task):
    
    def __init__(self, processor):
        Thread.__init__(self)
        self.Processor = processor
        
    def close_processor(self):
        self.Processor.join()
        self.Processor.close_output()
            
    def run(self):
        pass
        

class _OutputIterator(object):
    
    def __init__(self, processor, queue):
        self.Queue = queue
        self.Processor = processor
        
    def __next__(self):
        try:    out, exc_info = self.Queue.pop()
        except StopIteration:
            self.Processor._remove_output_queue(self.Queue)
            raise
        if exc_info:
            _, e, _ = exc_info
            raise e
        else:
            return out
            
    def __del__(self):
        self.Processor._remove_output_queue(self.Queue)
        
class Processor(Primitive):
    
    Default = ""

    def __init__(self, max_workers = None, queue_capacity = None, name=None, output = Default, stagger=None, delegate=None,
            put_timeout=None):
        Primitive.__init__(self, name=name)
        self.Output = DEQueue() if output is self.Default else output
        self.WorkerQueue = TaskQueue(max_workers, capacity=queue_capacity, stagger=stagger)
        self.Delegate = delegate
        self.PutTimeout = put_timeout
        self.Closed = False
        
    def hold(self):
        self.WorkerQueue.hold()

    def release(self):
        self.WorkerQueue.release()
        
    def open(self):
        self.Output.open()
        self.Closed = False

    @synchronized
    def close(self):
        self.Closed = True
        t = PyThread(target=self.wait_and_close_output)
        t.start()
    
    def wait_and_close_output(self):
        print("wait_and_close_output: joining...")
        self.join()
        print("          join done")
        self.Output.close()

    @synchronized
    def put(self, item, timeout=-1):
        if self.Closed:
            raise RuntimeError("Processor is closed")
        if timeout == -1: timeout = self.PutTimeout
        promise = Promise()
        self.WorkerQueue.addTask(_WorkerTask(self, item, promise), timeout)
        return promise

    def get(self):
        return self.Output.get()

    def join(self):
        return self.WorkerQueue.join()

    def _process(self, item, promise):
        #print("%x: Processor._process: item: %s" % (id(self), item))
        try:    
            out = self.process(item)
        except:
            exc_type, exc_value, tb = sys.exc_info()
            if self.Delegate is not None:
                self.Delegate.itemFailed(item, exc_type, exc_value, tb)
            raise
            promise.exception(exc_type, exc_value, tb)
            if self.OutputQueue is not None:
                self.OutputQueue.append((None, (exc_type, exc_value, tb)))
        else:
            #print("%x: out: %s" % (id(self), out))
            if self.Delegate is not None:
                self.Delegate.itemProcessed(item, out)
            if self.Output is not None and out is not None:
                    next_promise = self.Output.put(out)
                    if next_promise is not None:
                        # fulfil this promise later, when the output processor has done with the item
                        next_promise.Name = "secondary"
                        next_promise.chain(promise)
                        return
            promise.complete(out)

    def process(self, item):
        # override me
        pass
        
    def waitingTasks(self):
        return self.WorkerQueue.waitingTasks()
        
    def activeTasks(self):
        return self.WorkerQueue.activeTasks()
        
    def tasks(self):
        return self.WorkerQueue.tasks()
        
    def nrunning(self):
        return self.WorkerQueue.nrunning()
        
    def nwaiting(self):
        return self.WorkerQueue.nwaiting()
        
    def counts(self):
        return self.WorkerQueue.counts()
        
    @synchronized
    def __iter__(self):
        return iter(self.Output)



from .core import Primitive, synchronized
from .promise import Promise

class Barrier(Primitive):

    def __init__(self, height):
        Primitive.__init__(self)
        self.Queue = []
        self.Height = height
        
    def wait(self, timeout=None):
        with self:
            if len(self.Queue) >= self.Height - 1:
                self.flush()
                return
            else:
                promise = Promise()
                self.Queue.append(promise)
        promise.wait(timeout=timeout)
        
    @synchronized
    def flush(self):
        for p in self.Queue:
            p.complete()
        self.Queue = []
        



from threading import RLock, Thread, Event, Condition, Semaphore, currentThread, get_ident
import time
import sys

Waiting = []
In = []

class Timeout(Exception):
    pass

class QueueClosed(Exception):
    pass

def threadName():
    t = currentThread()
    return str(t)

def synchronized(method):
    def smethod(self, *params, **args):
        me = get_ident()
        with self:
            out = method(self, *params, **args)
        return out
    smethod.__doc__ = method.__doc__
    return smethod

def gated(method):
    def smethod(self, *params, **args):
        with self._Gate:
            out = method(self, *params, **args)
        return out
    smethod.__doc__ = method.__doc__
    return smethod


def printWaiting():
    print("waiting:----")
    for w in Waiting:
        print(w)
    print("in:---------")
    for w in In:
        print(w)

class UnlockContext(object):
    
    def __init__(self, prim):
        self.Prim = prim
        
    def __enter__(self):
        self.Prim._Lock.__exit__(None, None, None)
        
    def __exit__(self, exc_type, exc_value, traceback):
        self.Prim._Lock.__enter__()    

class Primitive:
    def __init__(self, gate=1, lock=None, name=None):
        """
        Initiaslizes new Primitive object
        
        Args:
            gate (int): initial value for the Gate semapthore. Default = 1
            lock (Lock or RLock): Lock or RLock for the primitive to use. Default - new RLock object
            name (str): Name for the primitive. Default - unnamed
        """
        self._Kind = self.__class__.__name__
        self._Lock = lock if lock is not None else RLock()
        #print ("Primitive:", self, " _Lock:", self._Lock)
        self._WakeUp = Condition(self._Lock)
        self._Gate = Semaphore(gate)
        self.Name = name
        self.Timer = None
        
    def __str__(self):
        ident = ('"%s"' % (self.Name,)) if self.Name else ("@%s" % (("%x" % (id(self),))[-4:],))
        return "[%s %s]" % (self._Kind, ident)

    def __get_kind(self):
        return self._Kind

    def __set_kind(self, kind):
        self._Kind = kind

    kind = property(__get_kind, __set_kind)

    def getLock(self):
        """Returns the primitive's lock object. It can be used to create Primitives using the same lock object.
        
        Returns:
            Lock or RLock: primitive's lock object
        """
        return self._Lock
        
    def __enter__(self):
        """Primitive's context entry. Using a primitve as a context is equivalent to using its
        lock as the context. The following are equivalent:
        
            with my_primitive:
                ...
        
            with my_primitive.getLock():
                ...
        """
        return self._Lock.__enter__()
        
    def __exit__(self, exc_type, exc_value, traceback):
        """Primitive's context exit
        """
        return self._Lock.__exit__(exc_type, exc_value, traceback)

    def can_lock(self):
        """Returns True if the thread can lock the Primitive's lock immediately, False otherwise.
        Does not guarantee that a subsequent attempt to lock the Primitive will succeed immediately.
        """
        if self._Lock.acquire(blocking=False):
            self._Lock.release()
            return True
        else:
            return False

    @property
    def unlock(self):
        """Creates a context which can be used to temporarily release the lock acquired on the
        primitive eariler. For example:
        
            with my_primitive:
                # do something, which needs to be done while the primitive is locked
                with my_primitive.unlock:
                    # temporarily unlock the primitive and do something else
                # re-acquire the primitive's lock again, possibly waiting for it
                # continue doing something while the primitive is locked again
        """
        return UnlockContext(self)

    @synchronized
    def sleep(self, timeout = None, function=None, arguments=()):
        """
        Blocks until wakep() method is called on the same primitive. The primitive will be unlocked for the duration of the sleep() call.
        Then the primitive will be locked again and if the function is provided, it will be called while the primitive is locked
        and then the sleep() methot will unlock the primitive and return the result.
        
        Args:
            timeout (float or int): time-out. If timed-out, RuntimeError exception will be raised.
        """
        self._WakeUp.wait(timeout)
        if function is not None:
            result = function(*arguments)
            return result

    @synchronized
    def sleep_until(self, predicate, *params, timeout = None, **args):
        """
        Blocks until a condition is satisfied. The method will continue calling sleep() and when it wakes up, it will call the
        predicate function and check if the predicate is satisfied and if not go back to sleep. The predicate function is
        expected to return True or False. Basically, sleep_until implements this loop inside a critical section.
        Note that the primitive is unlocked while the sleep() is in progress.
        
            with self:
                while not timed-out:
                    primitive.sleep()
                    if predicate(*params, **args):
                        break

        Args:
            predicate (function): a function returning True or False. The method will exit when the function returns True
            params: positional arguments to pass to each predicate function call
            args: keyword arguments to pass to each predicate function call
            timeout (int or float): timeout. If timed-out, RuntimeError exception will be raised
        """
        #print("sleep", self, get_ident(), "   condition lock:", self._WakeUp._lock, "...")
        t1 = None if timeout is None else time.time() + timeout
        while (t1 is None or time.time() < t1):
            if predicate(*params, **args):
                return
            delta = None
            if t1 is not None:
                delta = max(0.0, t1 - time.time())
            self.sleep(delta)
        else:
            raise Timeout()
            
    @synchronized
    def wakeup(self, n=1, all=True, function=None, arguments=()):
        """Wakes up all or some thread, which is sleeping on the primitive.
        
        Args:
            n (int): number of threads to wake up. Default is 1
            all (bool): wake up all threads sleeping on the primitive. If True, ``n`` argument is ignored
        """
        
        if function is not None:
            function(*arguments)
        if all:
            self._WakeUp.notifyAll()
        else:
            self._WakeUp.notify(n)
            
    @synchronized
    def alarm(self, *params, **args):
        """Starts a new "alarm" Timer thread associated with the Primitive. A Primitive can have only one alarm thread associated with it.
        If another alarm thread was already created using a previous call to alarm(), the old thread will be cancelled and
        deleted.
        
        Args:
            params: positional arguments to pass to the Timer constructor
            args: keyword arguments to pass to the Timer constructor
        """
        self.cancel_alarm()
        self.Timer = Timer(*params, **args)

    @synchronized
    def cancel_alarm(self):
        """
        Cancels the alarm thread associated with the Primitive, if any. 
        """
        
        if self.Timer is not None:
            self.Timer.cancel()
            self.Timer = None

class PyThread(Thread, Primitive):
    def __init__(self, *params, name=None, **args):
        """Initializes a new PyThread object. PyThread is a subclass of both threading.Thread and Primitive, so it combines features of both.
        
        Args:
            name (string):  Name for the Primitive
            params: positional arguments for threading.Thread constructor
            args: keyword arguments for threading.Thread constructor
        """
        Thread.__init__(self, *params, **args)
        Primitive.__init__(self, name=name)
        self.Stop = False
        
    def stop(self):
        self.Stop = True

class Timer(PyThread):
    
    def __init__(self, fcn, *params, t=None, interval=None, start=True, name=None, daemon=True, onexception=None, **args):
        """Initializes new Timer thread. pythreader's Timer is similar in functionality with threading.Timer, but it has additional
        functionality. In particular pythreader's Timer can run as a periodic timer, firing at specified frequency until it is cancelled.
        
        Args:
            fcn (function): function to call every time the Timer fires
            params: positional arguments to pass to the ``fcn``
            args: keyword arguments to pass to the ``fcn``
            t (numeric): specifies the time when to fire the Timer first time. ``t`` can be either absolute timestamp - time in seconds since the
                Epoch or relative to current time. If ``t`` is less than 3e8 (~10 years), then it is interpreted as relative to current time.
                Default: current time plus ``interval``
            start (bool): to start the Timer thread immediately. Otherwise, the Timer thread will be created but not run until started explicitly
                using ``start()`` method. Default: start immediately
            daemon (bool): whether the Timer thread should be a daemon thread. Default: True
            onexception (function): a callback to call if ``fcn`` raies an exception. The callback will be called with 3 arguments:
                exception type, exception value and the traceback, similar to what sys.exc_info() returns. Default: ignore any exceptions raised by ``fcn``
        """
        
        PyThread.__init__(self, name=name, daemon=daemon)
        if t is None:
            self.T = time.time() + interval
        else:
            self.T = t if t > 3e8 else time.time() + t
        self.Fcn = fcn
        self.OnException = onexception
        self.Params = params
        self.Args = args
        self.Interval = interval
        self.Cancelled = False
        self.Paused = False
        if start:
            self.start()

    def run(self):
        try:
            again = True
            while again and not self.Cancelled:
                again = False
                now = time.time()
                if now < self.T:
                    self.sleep(self.T - now)
                while self.Paused and not self.Cancelled:
                    self.sleep(100)
                if not self.Cancelled:
                    try:    self.Fcn(*self.Params, **self.Args)
                    except:
                        if self.OnException is not None:
                            try:
                                self.OnException(*sys.exc_info())
                            except:
                                pass
                    if not self.Cancelled and self.Interval:
                        self.T = time.time() + self.Interval
                        again = True
        finally:
            # to break any circular links
            self.Fcn = self.Args = self.Params = None

    def cancel(self):
        """Cancels the Timer thread
        """
        self.Cancelled = True
        self.wakeup()

    def pause(self):
        """Pauses the Timer thread. In fact the thread will continue running, but it will not fire intil ``resume`` method is called
        """
        self.Paused = True

    def resume(self):
        """Resumes the timer firing. Next time the timer will not necessarily fire immediately when resumed, but it will fire at the next
        end of the firing interval.
        """
        self.Paused = False
        self.wakeup()





from .core import PyThread
from .dequeue import DEQueue

class Producer(PyThread):
    
    def __init__(self, capacity=None, name=None):
        PyThread.__init__(self, name=name)
        self.OutQueue = DEQueue(capacity=capacity)
    
    def __iter__(self):
        try:    self.start()
        except RuntimeError:
            pass    # already started: ignore
        for item in self.OutQueue:
            yield item
            
    def emit(self, item):
        self.OutQueue.append(item)
        
    def close(self):
        self.OutQueue.close()




from pythreader import Task, TaskQueue
import sys, traceback

class TaskOK(Task):
    
    def run(self):
        print(">>> TaskOK: OK")
        
    def __str__(self):
        return "Task<OK>"
        
class TaskError(Task):
    
    def run(self):
        print(">>> TaskError: generating exception")
        raise ValueError("Generated ValueError")
        
    def __str__(self):
        return "Task<Error>"
        

class Delegate(object):

    def taskFailed(self, queue, task, etype, evalue, etb):
        print(f"Delegate: exception reported in task {task}")
        traceback.print_exception(etype, evalue, etb)
        
    def taskEnded(self, queue, task):
        print(f"Delegate: task ended: {task}")
        
        
d = Delegate()

q1 = TaskQueue(2)
q2 = TaskQueue(2, delegate = d)
tok = TaskOK()
terr = TaskError()


print("-------- Queue without delegae --------")
for _ in range(100):
    q1 << tok << terr
q1.join()

print("-------- Queue with delegae --------")
for _ in range(100):
    q2 << tok << terr
q2.join()





        



from .core import Primitive, synchronized
import time

class DEQueue(Primitive):

    def __init__(self, capacity=None):
        Primitive.__init__(self)
        self.Capacity = capacity
        self.List = []
        self.Closed = False
    
    @synchronized
    def close(self):
        self.Closed = True
        self.wakeup()

    @synchronized
    def open(self):
        self.Closed = False
        self.wakeup()

    def _wait_for_room(self, timeout):
        # must be called from a synchronized method !
        t0 = time.time()
        t1 = None if timeout is None else t0 + timeout
        while self.Capacity is not None and len(self.List) >= self.Capacity \
                        and not self.Closed:
            dt = None
            if t1 is not None:
                t = time.time()
                if t > t1:
                    raise RuntimeError("Operation timed-out")
                dt = t1 - t
            self.sleep(dt)

    @synchronized    
    def append(self, item, timeout=None, force=False):
        if not force:
            self._wait_for_room(timeout)
        if self.Closed:
            raise RuntimeError("Queue is closed")
        self.List.append(item)
        self.wakeup()
        
    def __lshift__(self, item):
        return self.append(item)
    
    put = append
    
    @synchronized    
    def insert(self, item, timeout=None, force=False):
        if not force:
            self._wait_for_room(timeout)
        if self.Closed:
            raise RuntimeError("Queue is closed")
        self.List.insert(0, item)
        self.wakeup()

    def __rrshift__(self, item):
        return self.insert(item)

    @synchronized
    def pop(self, index=0, timeout=None):
        while not (self.List or self.Closed):
            self.sleep(timeout)
        try:    
            item = self.List.pop(index)
            self.wakeup()       # in case someone is waiting to add an item
        except IndexError:
            item = None         # closed
        return item

    get = pop

    #
    # Iterator protocol
    # 
    # for item in queue:        # wait for next item to arrive
    #   # ... process item
    #
    
    def __iter__(self):
        return self

    def __next__(self):
        item = self.pop()
        if item is None:
            raise StopIteration()
        else:
            return item
        raise StopIteration()

    next = __next__
        
    @synchronized
    def flush(self):
        self.List = []
        self.wakeup()
        
    @synchronized
    def items(self):
        return self.List[:]
        
    @synchronized
    def look(self):
        return self.List[0] if self.List else None
        
    @synchronized
    def popIfFirst(self, x):
        if self.List:
            first = self.List[0]
            if first is x or first == x:
                self.pop()
                return x
        return None

    def empty(self):
        return not self.List
        
    def __len__(self):
        return len(self.List)
    
    def __contains__(self, item):
        return item in self.List
        
    def remove(self, item):
        self.List.remove(item)
        self.wakeup()



from pythreader import Processor
import random, time

class P1(Processor):
    
    def process(self, x):

        time.sleep(random.random()*0.5)
        return x+1
        
class P2(Processor):
    
    def process(self, x):
        time.sleep(random.random()*0.5)
        return x**2
        

p2 = P2(3, name="P2")
p1 = P1(3, output=p2, name="P1")

promises = [p1.put(x) for x in range(10)]

if False:
    for pr in promises:
        print(pr.Data, "->", pr.wait())
i = 0
for y in p2:
	print(y)
	i += 1
	if i > 3:
		break
print("reopened")
for y in p2:
        print(y)




import random, time, uuid
from pythreader import Processor, PyThread


class MyProcessor(Processor):
    """
    Simple processor
    """

    def process(self, item):
        return "%s -> %s" % (item, item[::-1])

processor = MyProcessor()

# test it quickly
processor.put("hello!")
print("Simple hello test output:", processor.get())

class Producer(PyThread):
    """
    Producer which will feed our processor with some items to process
    """
    def __init__(self, processor, n, sleep_between_items):
        PyThread.__init__(self)
        self.Processor = processor
        self.N = n
        self.Delay = sleep_between_items
    
    def run(self):
        for _ in range(self.N):
            if self.Delay:
                time.sleep(random.random())
            self.Processor.put(uuid.uuid4().hex[:4])
        self.Processor.close()
        print(f"producer: {self.N} items sent. Processor closed")

for run in range(1, 4):
    print("\n--- Starting run", run, "---")
    processor.open()
    producer = Producer(processor, 10, run%2)
    producer.start()
    n = 0
    for out in processor:
        print(out)
        n += 1
    print(f"producer output ended. {n} items received")
    producer.join()
    print("--- Ended or run", run, "---")







from .core import Primitive, synchronized
from .promise import Promise
from .dequeue import DEQueue
from .task_queue import Task, TaskQueue

class Gate(Primitive):

    def __init__(self, open=False):
        Primitive.__init__(self)
        self.Queue = []
        self.Open = open
        
    def wait(self, timeout=None):
        with self:
            if self.Open:   return
            promise = Promise()
            self.Queue.append(promise)
        return promise.wait(timeout=timeout)
        
    @synchronized
    def open(self, message=None):
        self.Open = True
        for p in self.Queue:
            p.complete(message)
        self.Queue = []

    @synchronized
    def close(self):
        self.Open = False
        
    @synchronized
    def pulse(self, message=None, limit=None):
        #
        # unblocks first or last "limit" blocked processes keeping the gate closed
        # if limit is None, unblocks all waiting processes
        # if limit > 0, unblocks first "limit" processes
        # if limit < 0, unblocks last "limit" processes
        #
        if limit is None:
            to_release = self.Queue
            self.Queue = []
        elif limit < 0:
            to_release = self.Queue[limit:]
            self.Queue = self.Queue[:limit]
        else:
            to_release = self.Queue[:limit]
            self.Queue = self.Queue[limit:]
        for p in to_release:
            p.complete(message)

            
            
        
    
        
        
        


from .core import Primitive, synchronized, Timeout
from threading import get_ident, RLock

class DebugLock(object):
    
    def __init__(self):
        self.R = RLock()
        
    def acquire(self, *params, **args):
        print(self, "acquire by ", get_ident(), "...")
        self.R.acquire(*params, **args)
        print(self, "acquired by ", get_ident())
        
    def release(self):
        print(self, "released by ", get_ident())
        self.R.release()
        
    def __enter__(self):
        return self.acquire()
        
    def __exit__(self, *params):
        return self.release()
    

class Promise(Primitive):
    
    def __init__(self, data=None, callbacks = [], name=None):
        """
        Creates new Promise object.
        
        Args:
            data: Arbitrary user-defined data to be associated with the promise. Can be anything. pythreader does not use it.
            callbacks (list): List of Promise callbacks objects. Each object may have ``oncomplete`` and/or ``onexception`` methods defined. See Notes below.
            name (string): Name for the new Promise object, optional.
        
        Notes:
            
        """
        Primitive.__init__(self, name=name)    #, lock=DebugLock())
        self.Data = data
        self.Callbacks = callbacks[:]
        self.Complete = False
        self.Cancelled = False
        self.Result = None
        self.ExceptionInfo = None     # or tuple (exc_type, exc_value, exc_traceback)
        self.RaiseException = True
        self.OnComplete = self.OnException = self.OnCancel = None
        self.Chained = []

    @synchronized
    def then(self, oncomplete=None, onexception=None):
        """
        Creates new Promise object, which will be chained to the ``self`` promise with provided ``oncomplete`` and/or ``onexception`` callbacks.
        
        Args:
            oncomplete (function): the new promise completion callback function
            oncexception (function): the new promise exception callback function
        """
        p = Promise()
        if oncomplete is not None:
            p.oncomplete(oncomplete)
        if onexception is not None:
            p.onexception(onexception)
        self.chain(p)
        return p

    @synchronized
    def oncomplete(self, cb):
        """Sets promise completion callback. If the promise has already completed, the callback function is called immediately.
        Old completion callback is removed.

        Args:
            cb (function):    Function to be called when the promise completes. The function will be called with the following arguments:

                cb(promise, result)
        """
        if self.Complete:
            cb(self, self.Result)
        self.OnComplete = cb
        return self
        
    @synchronized
    def onexception(self, cb):
        """Sets promise exception callback. If the promise has failed already, the callback function will be called immediately.
        Old exception callback is removed.
        
        Args:
            cb (function):    Function to be called when the promise fails. The function will be called with the following arguments:
        
                cb(promise, exception_type, exception_value, traceback)
        """
        if self.ExceptionInfo:
            cb(self, *self.ExceptionInfo)
        self.OnException = cb
        return self
        
    @synchronized
    def oncancel(self, cb):
        """Sets promise cancellation callback. If the promise has been cancelled already, the callback function will be called immediately.
        Old cancellation callback is removed.
        
        Args:
            cb (function):    Function to be called when the promise is cancelled. The function will be called with the promise as the only
                argiment.
        """
        if self.Cancelled:
            cb(self)
        self.OnCancel = cb
        return self
        
    @synchronized
    def addCallback(self, cb):
        """Adds a new promise callback object to the promise callback list. If the promise has completed, failed or cancelled already,
            the new callback object will be notified immediately and will not be added to the list.
        
        Args:
            cb (object):    New Promise callback object. The callback object is expected to have some (or none) of the following methods:
        
                def oncancel(self, promise):
                    # the promise was cancelled

                def oncomplete(self, promise, result):
                    # the promise completed

                def onexception(self, promise, exception_type, exception_value, traceback):
                    # the promise failed        
        """
        if self.Cancelled and hasattr(cb, "oncancel"):
            cb.oncancel(self)
        elif self.Complete and hasattr(cb, "oncomplete"):
            cb.oncomplete(self, self.Result)
        elif self.ExceptionInfo and hasattr(cb, "onexception"):
            cb.onexception(self, *self.ExceptionInfo)
        else:
            self.Callbacks.append(cb)
        return self

    @synchronized
    def chain(self, *promises, cancel_chained = True):
        """Adds new chained promises to ``self``. If the ``self`` promise has delivered (completed or failed) already, the new
        promises will be delivered with the same status immediately. If the ``self`` promise has been cancelled, the new promises will be cancelled too.
        
        Args:
            promises (list of Promises): list of promises to add as chained
            cancel_chained (bool): whether to cancel the new promises if the ``self`` promise has been cancelled. Default = True
        """
        if self.ExceptionInfo:
            exc_type, exc_value, exc_traceback = self.ExceptionInfo
            for p in promises:
                p.exception(exc_type, exc_value, exc_traceback)
        elif self.Complete:
            for p in promises:
                p.complete(self.Result)
        elif self.Cancelled and cancel_chained:
            for p in promises:
                p.cancel()
        else:
            self.Chained += list(promises)
        return self

    @synchronized
    def complete(self, result=None):
        """Delivers the promise as successfully completed with given result
            
            Args:
                result: the promise result to be delivered
        """
        self.Result = result
        self.Complete = True
        if not self.Cancelled:
            stop = False
            if self.OnComplete is not None:
                stop = not not self.OnComplete(self, self.Result)
            for cb in self.Callbacks:
                if stop:
                    break
                if hasattr(cb, "oncomplete"):
                    stop = not not cb.oncomplete(self.Result, self)
        for p in self.Chained:
            p.complete(result)
        self.wakeup()
        self._cleanup()

    @synchronized
    def exception(self, exc_type, exc_value, exc_traceback):
        """Fails the promise with the exception
            
            Args:
                exc_type: exception type
                exc_value: exception value
                exc_traceback: exception traceback
        """
        self.ExceptionInfo = (exc_type, exc_value, exc_traceback)
        if not self.Cancelled:
            stop = False
            if self.OnException is not None:
                stop = bool(self.OnException(self, exc_type, exc_value, exc_traceback))
            for cb in self.Callbacks:
                if stop:
                    break
                if hasattr(cb, "onexception"):
                    stop = not not cb.onexception(self, exc_type, exc_value, exc_traceback)
            self.RaiseException = not stop
        for p in self.Chained:
            #print("forwarding exception to the chained promise...")
            p.exception(exc_type, exc_value, exc_traceback)
        self.wakeup()
        self._cleanup()

    @synchronized
    def cancel(self, cancel_chained = True):
        """Cancels the promise
        
        Args:
            cancel_chained (bool): whether to cancel chained promises too. Default = True
        """
        if not self.Cancelled:
            stop = False
            if self.OnCancel is not None:
                #print("calling OnException...")
                stop = not not self.OnCancel(self)
            for cb in self.Callbacks:
                if stop:
                    break
                if hasattr(cb, "oncancel"):
                    stop = not not cb.oncancel(self)
            if cancel_chained:
                for p in self.Chained:
                    p.cancel()
        self.Cancelled = True
        self.wakeup()
        self._cleanup()

    @synchronized
    def wait(self, timeout=None):
        """Blocks until the promise closes (completes, fails or gets cancelled).
        
        Args:
            timeout (numeric): Time-out in seconds. If the operation times out, the RuntimeError exception will be raised.
        
        Returns:
            Object: promise result, if the promise completes successfully. If the promsie is cancelled, None is returned.
        
        Raises:
            If the promise fails with an exception and the exception was not handled by one of the promise's callbacks,
            then the original exception will be raised.
        """
        
        #print("thread %s: wait(%s)..." % (get_ident(), self))
        pred = lambda x: x.Complete or x.Cancelled or self.ExceptionInfo is not None
        self.sleep_until(pred, self, timeout=timeout)
        try:
            if self.Complete:
                return self.Result
            elif self.Cancelled:
                return None
            elif self.ExceptionInfo:
                if self.RaiseException:
                    _, e, tb = self.ExceptionInfo
                    raise e.with_traceback(tb)
            else:
                raise Timeout()
        finally:
            self._cleanup()

    def _cleanup(self):
        self.Chained = []
        self.Callbacks = []
        self.OnException = self.OnComplete = None

    @staticmethod
    def all(*args):
        """Static method to create an ``ANDPromise`` - a promise-like object, which represents completion of all the argument promsies. 
        
        Args:
            *args: promise objects to combine
        """
        
        if args and isinstance(args[0], (tuple, list)):
            promises = args[0]
        else:
            promises = args
        return ANDPromise(promises)

    @staticmethod
    def any(*args):
        """Static method to create an ``ORPromise`` - a promise-like object, which represents completion of one of the argument promsies. 
        
        Args:
            *args: promise objects to combine
        """
        if args and isinstance(args[0], (tuple, list)):
            promises = args[0]
        else:
            promises = args
        return ORPromise(promises)
    
class ORPromise(Primitive):
    
    def __init__(self, promises):
        Primitive.__init__(self)
        self.Fulfilled = None
        for p in promises:
            p.addCallback(self)

    @synchronized
    def oncomplete(self, promise, result):
        if self.Fulfilled is None:
            self.Fulfilled = promise
            self.wakeup()

    @synchronized
    def wait(self, timeout = None):
        while self.Fulfilled is None:
            self.sleep(timeout)
        return self.Fulfilled.Result

class ANDPromise(Primitive):
    
    def __init__(self, promises):
        Primitive.__init__(self)
        self.Promises = promises
        
    def wait(self, timeout=None):
        return [p.wait(timeout) for p in self.Promises]



from .core import PyThread
from .promise import Promise
import sys, os, subprocess
from subprocess import Popen

def to_str(b):
    if isinstance(b, bytes):
        b = b.decode("utf-8")
    return b

def to_bytes(b):
    if isinstance(b, str):
        b = b.encode("utf-8")
    return b

class ShellCommand(PyThread):

    def __init__(self, command, promise_data=None, cwd=None, env=None, input=None, daemon=None, timeout=None, name=None):
        PyThread.__init__(self, daemon=daemon, name=name)
        self.Command = command
        self.CWD = cwd
        self.Env = env
        self.Input = input
        self.Timeout = timeout
        self.TimedOut = False
        self.Promise = Promise(promise_data)
        self.Out = None
        self.Err = None
        self.Returncode = None

    def run(self):
        p = Popen(self.Command, shell=True,
                    stdin=subprocess.DEVNULL if self.Input is None else suprocess.PIPE,
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                    cwd=self.CWD, env=self.Env)
        try:    
            out, err = p.communicate(self.Input, self.Timeout)
        except subprocess.TimeoutExpired:
            self.TimedOut = True
            p.kill()
            out, err = p.communicate()
            self.Promise.exception(RuntimeError, RuntimeError("time-out"), None)
            raise RuntimeError("time-out")
        self.Out = to_str(out)
        self.Err = to_str(err)
        self.Returncode = p.returncode
        self.Promise.complete((p.returncode, self.Out, self.Err))
        return self.Returncode, self.Out, self.Err

    @staticmethod
    def execute(command, cwd=None, env=None, input=None, timeout=None, daemon=True):
        s = ShellCommand(command, cwd=cwd, env=env, input=input, daemon=daemon, timeout=timeout)
        return s.run()


import threading
from pythreader import RWLock

if __name__ == '__main__':
    import random, time
    
    class T(threading.Thread):
        
        Interlock = threading.RLock()
        
        @classmethod
        def printStats(cls, lock):
            with cls.Interlock:
                exclusive, shared = lock.owners()
                print (exclusive, len(shared))
                
        
        def __init__(self, lock):
            threading.Thread.__init__(self)
            self.Lock = lock
        
        def run(self):
            while True:
                if random.random() < 0.2:
                    with self.Lock.exclusive:
                        print ("exclusive")
                        self.printStats(self.Lock)
                        time.sleep(random.random() * 5.0)
                        self.printStats(self.Lock)
                else:
                    with self.Lock.shared:
                        print ("shared")
                        self.printStats(self.Lock)
                        time.sleep(random.random() * 5.0)
                        self.printStats(self.Lock)
                    
    l = RWLock()
    nt = 10
    threads = [T(l) for _ in range(nt)]
    for t in threads:
        t.start()
        
    for t in threads:
        t.join()
    
    
    
    
    



from pythreader import Processor, Flag
import random
from threading import Condition

done = Flag()

class MyProcessor(Processor):
    
    def __init__(self, name, nworkers, output = None):
        Processor.__init__(self, nworkers, output = output, name=name)

    def process(self, lst):
        lst.append(self.Name)
        if random.random() < 0.2:
            done.value = "done"
            return None
        else:
            return lst

p1 = MyProcessor("A", 2)
p2 = MyProcessor("B", 2, output = p1)
p1.Output = p2

lst = []
p1.put(lst)

done.sleep_until("done")

print("Result:", lst)


import time, traceback, sys
from datetime import datetime, timedelta
from .core import Primitive, PyThread, synchronized
from .dequeue import DEQueue
from .promise import Promise
from threading import Timer

class TaskQueueDelegate(object):
    
    # abstract class
    
    def taskIsStarting(self, queue, task):
        pass

    def taskStarted(self, queue, task):
        pass

    def taskEnded(self, queue, task, result):
        pass

    def taskFailed(self, queue, task, exc_type, exc_value, tback):
        pass
        
def _after_time(after):
    if after is None:   return None
    if isinstance(after, timedelta):
        after = time.time() + after.totalseconds()
    elif isinstance(after, datetime):
        after = after.timestamp()
    if after < 10*365*24*3600:  # 1980
        after = time.time() + after
    return after
    
def _time_interval(interval):
    if isinstance(interval, timedelta):
        interval = interval.totalseconds()
    return interval
        
class Task(Primitive):

    class _TaskPrivate(object):
        pass

    def __init__(self, name=None):
        Primitive.__init__(self, name=name)
        self.Created = time.time()
        self.Queued = None
        self.Started = None
        self.Ended = None
        # Use _Private member to avoid name clashes with a subclass
        self._Private = self._TaskPrivate()
        self._Private.Promise = None
        self._Private.RepeatInterval = None
        self._Private.RunCount = 1
        self._Private.After = None
        self._Private.Running = False               # True actually means that the Executor thread was created and about to be started
        self._Private.LastStart = None
        self._Private.Cancelled = False

    def __repr__(self):
        return str(self)

    @property
    def promise(self):
        """
        Returns:
            Promise: the Promise associated with the Task
        """
        return self._Private.Promise
        
    @synchronized
    def deliver_promise(self, result=None):
        promise = self._Private.Promise
        if promise is not None:
            if self.is_cancelled:
                promise.cancel()
            else:
                promise.complete(result)
            self._Private.Promise = None

    @synchronized
    def cancel(self):
        """
        Cancels the task. If the task is alreadt running, it will not be interrupted. Otherwise, it will be removed from any
        TaskQueue the task was added to. The task promise will be cancelled.
        """
        if not self._Private.Cancelled:
            self._Private.Cancelled = True
            promise = self.promise
            if promise is not None:
                promise.cancel()

    @property
    def is_cancelled(self):
        """
        Returns:
            boolean: whether the task is cancelled
        """
        return self._Private.Cancelled

    @synchronized
    def repeat(self, after=None, count=1, interval=0):
        if not self._Private.Cancelled:
            self._Private.RunCount = count
            self._Private.RepeatInterval = _time_interval(interval)
            self._Private.After = _after_time(after)

    def run(self):
        raise NotImplementedError

    @property
    def has_started(self):
        """
        Returns:
            boolean: whether the task has started. For repeating tasks, the value indicates that the task was started at least once,
            so once the task starts first time, this propery will remain true even if the task is not running at the moment.
        """
        return self.Started is not None

    @property
    def is_running(self):
        """
        Returns:
            boolean: whether the task is currently running. This is not exactly the same as the ``running`` status of the corresponding
            thread, but very close. Strictly speaking, True value indicates that the task thread was created and is actually
            running or about to start.
        """
        return self._Private.Running

    @synchronized
    def to_be_repeated(self):
        interval = self._Private.RepeatInterval
        count = self._Private.RunCount
        repeat = interval is not None and (count is None or count > 0) \
            or interval is None and count is not None and count > 0
        return repeat and self.should_repeat()

    @property
    @synchronized
    def has_ended(self):
        """
        Returns:
            boolean: whether the task has ended and will not re-start again.
        """
        return self.Started and not self.is_running and not self.to_be_repeated()

    @synchronized
    def _started(self):
        t = time.time()
        if self.Started is None:    self.Started = t
        if self._Private.RunCount is not None:
            self._Private.RunCount -= 1
        self._Private.LastStart = t
        self._Private.LastEnd = None

    @synchronized
    def _ended(self):
        self._Private.LastEnd = self.Ended = time.time()

    def _queued(self):
        self.Queued = time.time()

    def __rshift__(self, queue):
        if not isinstance(queue, TaskQueue):
            raise TypeError(f"unsupported operand type(s) for >>: 'Task' and %s" % (type(queue),))
        return queue.insert(self)

    # overridable
    
    def should_repeat(self):
        """Overridable method. For repeating tasks, the Task Queue will call this method right after the task finishes
        and it is repeatable. 
        If the method returns False, the task fill not be repeated even if its run count is still greater than 0.

        Returns:
            boolean: whether the task should be repeated
        """
        return True

class FunctionTask(Task):

    def __init__(self, fcn, *params, **args):
        Task.__init__(self)
        self.F = fcn
        self.Params = params
        self.Args = args
        
    def run(self):
        result = self.F(*self.Params, **self.Args)
        #self.F = self.Params = self.Args = None
        return result
        
class TaskQueue(Primitive):
    
    class ExecutorThread(PyThread):
        def __init__(self, queue, task):
            PyThread.__init__(self, daemon=True)
            self.Queue = queue
            self.Task = task
            task._Private.Running = True
            
        def run(self):
            task = self.Task
            task._started()             # this will decrement RunCount
            repeat = False
            try:
                if callable(task):
                    result = task()
                else:
                    result = task.run()
                task._ended()
                #print(task._Private.__dict__)
                repeat = task.to_be_repeated() \
                    and self.Queue.taskWillRepeat(task, result, task._Private.After, task._Private.RunCount) is not False
                #print("repeat:", repeat)
                if repeat:
                    interval = task._Private.RepeatInterval or 0
                    task._Private.After = (task._Private.LastStart if task._Private.After is None else task._Private.After) + interval
                else:
                    task.deliver_promise(result)
                    self.Queue.taskEnded(task, result)
            except:
                exc_type, value, tb = sys.exc_info()
                traceback.print_exc()
                task._ended()
                promise = task.promise
                if promise is not None:
                    promise.exception(exc_type, value, tb)
                self.Queue.taskFailed(self.Task, exc_type, value, tb)
            finally:
                task._Private.Running = False
                self.Queue.threadEnded(task, repeat)
                self.Queue = None

    def __init__(self, nworkers=None, capacity=None, stagger=0.0, tasks = [], delegate=None, 
                        name=None):
        """Initializes the TaskQueue object
        
        Args:
            nworkers (int): maximum number of tasks to be executed concurrently. Default: no limit.
            capacity (int): maxinum number of tasks allowed in the queue before they start. If the capacity is reached, append() and insert() methods
                        will block. Default: no limit.

        Keyword Arguments:
            stagger (int or float): time interval in seconds between consecutive task starts. Default=0, no staggering.
            tasks (list of Task objects): initial task list to be added to the queue
            delegate (object): an object to receive callbacks with task status updates. If None, updates will not be sent.
            name (string): primitive name
            daemon (boolean): Threading daemon flag for the queue internal thread. Default = True
        """
        Primitive.__init__(self, name=name)
        self.NWorkers = nworkers
        self.Queue = DEQueue(capacity)
        self.Held = False
        self.Stagger = stagger
        self.LastStart = 0.0
        self.StartTimer = None
        self.Delegate = delegate
        self.Stop = False
        for t in tasks:
            self.addTask(t)
        
    def stop(self):
        """Stops the queue. Any attempt to add any new tasks will cause an exception. All running
        tasks will continue running, but new tasks will not start."""
        self.Stop = True
        self.Queue.close()
        self.cancel_alarm()
        
    def __add(self, mode, task, *params,
            timeout=None, promise_data=None, force=False,
            count = None, interval = None, after=None,
            **args):

        if interval is None and count is None:
            count = 1
        
        if not isinstance(task, Task):
            if callable(task):
                task = FunctionTask(task, *params, **args)
            else:
                raise ArgumentError("The task argument must be either a callable or a Task subclass instance")

        task._Private.Promise = promise = Promise(data=promise_data)

        task._Private.RunCount = count
        task._Private.RepeatInterval = _time_interval(interval)
        task._Private.After = _after_time(after)

        if mode == "insert":
            self.Queue.insert(task, timeout = timeout, force=force)
        else:           # mode == "append"
            self.Queue.append(task, timeout = timeout, force=force)
        task._queued()
        self.start_tasks()
        return task

    def reinsert_task(self, task):
        self.Queue.insert(task, force=True)

    def append(self, task, *params, timeout=None, promise_data=None, after=None, force=False, 
                count=None, interval=None, **args):
        """Appends the task to the end of the queue. If the queue is at or above its capacity, the method will block.
        
        Args:
            task (Task): A Task subclass instance to be added to the queue

        Keyword Arguments:
            timeout (int or float or timedelta): time to block if the queue is at or above the capacity. Default: block indefinitely.
            promise_data (object): data to be associated with the task's promise
            after (int or float or datetime): time to start the task after. 
                If ``after`` is numeric and < 365 days or it is a datetime.timedelta object, it is interpreted as time relative to the current time.
                Default: start as soon as possible
            force (boolean): ignore the queue capacity and append the task immediately. Default: False
            interval (numeric or datetime.timedelta): interval at which to repeat the task. Default: None
            count (int): how many times to repeat the task. Default None.
        
        Returns:
            Task: the task added to the queue. If the first argument was a callable, then the method will return a Task
                created for the callable and return it.

        Raises:
            RuntimeError: the queue is closed or the timeout expired
        """
        return self.__add("insert", task, *params, 
                after=after, timeout=timeout, promise_data=promise_data, force=force, count=count, interval=interval, **args)
        
    add = addTask = append
        
    def __iadd__(self, task):
        return self.addTask(task)

    def insert(self, task, *params, timeout = None, promise_data=None, after=None, force=False, count=None, interval=None, **args):
        """Inserts the task at the beginning of the queue. If the queue is at or above its capacity, the method will block.
           A Task can be also inserted into the queue using the '>>' operator. In this case, '>>' operator returns
           the promise object associated with the task: ``promise = task >> queue``.
        
        Args:
            task (Task): A Task subclass instance to be added to the queue or a callable

        Keyword Arguments:
            timeout (int or float or timedelta): time to block if the queue is at or above the capacity. Default: block indefinitely.
            promise_data (object): data to be associated with the task's promise
            after (int or float or datetime): time to start the task after. Default: start as soon as possible
                If ``after`` is numeric and < 365 days or it is a datetime.timedelta object, it is interpreted as time relative to the current time.
                Default: start as soon as possible
            force (boolean): ignore the queue capacity and append the task immediately. Default: False
            interval (numeric or datetime.timedelta): interval at which to repeat the task. Default: None
            count (int): how many times to repeat the task. Default None.
        
        Returns:
            Task: the task added to the queue. If the first argument was a callable, then the method will return a Task
                created for the callable and return it.
        
        Raises:
            RuntimeError: the queue is closed or the timeout expired
        """
        return self.__add("insert", task, *params, 
                after=after, timeout=timeout, promise_data=promise_data, force=force, count=count, interval=interval, **args)
        
    insertTask = insert

    def __lshift__(self, task):
        """Allows to append the task using the '<<' operator: ``promise = queue << task``.
        
        Returns:
            Promise: promise object associated with the task. The Promise will be delivered when the task ends.
        """
        return self.append(task)

    @synchronized
    def start_tasks(self):
        self.cancel_alarm()
        # remove cancelled tasks
        for t in self.Queue.items():
            if t.is_cancelled:
                self.Queue.remove(t)
        again = True
        while not (self.Stop or self.Held) and again:
            again = False
            now = time.time()
            if self.Stagger is not None and self.LastStart + self.Stagger > now:
                self.alarm(self.start_tasks, t = self.LastStart + self.Stagger)
            elif self.Queue:
                nrunning = self.nrunning()
                if (self.NWorkers is None or nrunning < self.NWorkers):
                    next_task = None
                    sleep_until = None
                    waiting_tasks = self.waitingTasks()
                    for t in self.waitingTasks():
                        after = t._Private.After
                        if after is None or after <= now:
                            next_task = t
                            break
                        else:
                            sleep_until = after if sleep_until is None else min(sleep_until, after)
                    if next_task is not None:
                        t = self.ExecutorThread(self, next_task)
                        t.kind = "%s.task" % (self.kind,)
                        self.LastStart = time.time()
                        self.call_delegate("taskIsStarting", self, next_task, t)
                        t.start()
                        self.call_delegate("taskStarted", self, next_task, t)
                        again = True
                    elif sleep_until is not None:
                        self.alarm(self.start_tasks, t=sleep_until)

    def threadEnded(self, task, repeat):
        if not repeat:
            try:    self.Queue.remove(task)
            except ValueError:  pass
            self.wakeup()               # in case someone is waiting for the queue to be drained
        self.start_tasks()
        
    def call_delegate(self, cb, *params):
        if self.Delegate is not None and hasattr(self.Delegate, cb):
            try:    
                return getattr(self.Delegate, cb)(*params)
            except:
                traceback.print_exc(file=sys.stderr)
            
    def taskEnded(self, task, result):
        return self.call_delegate("taskEnded", self, task, result)
        
    def taskWillRepeat(self, task, result, next_t, count):
        return self.call_delegate("taskWillRepeat", self, task, result, next_t, count)
        
    def taskFailed(self, task, exc_type, exc_value, tb):
        return self.call_delegate("taskFailed", self, task,  exc_type, exc_value, tb)
            
    def waitingTasks(self):
        """
        Returns:
            list: the list of tasks waiting in the queue
        """
        return [t for t in self.Queue.items() if not t.is_running]
        
    def activeTasks(self):
        """
        Returns:
            list: the list of running tasks
        """
        return [t for t in self.Queue.items() if t.is_running]
        
    @synchronized
    def tasks(self):
        """
        Returns:
            tuple: (self.waitingTasks(), self.activeTasks())
        """
        tasks = self.Queue.items()
        return [t for t in tasks if not t.is_running], [t for t in tasks if t.is_running]
        
    @synchronized
    def nrunning(self):
        """
        Returns:
            int: number of runnign tasks
        """
        return sum(t.is_running for t in self.Queue.items())
        
    @synchronized
    def nwaiting(self):
        """
        Returns:
            int: number of waiting tasks
        """
        return sum(not t.is_running for t in self.Queue.items())
        
    @synchronized
    def counts(self):
        """
        Returns:
            tuple: (self.nwaiting(), self.nrunning())
        """
        nrunning, nwaiting = 0, 0
        for t in self.Queue.items():
            if t.is_running:    nrunning += 1
            else:               nwaiting += 1
        return nwaiting, nrunning
    
    def hold(self):
        """
        Holds the queue, preventing new tasks from being started
        """
        self.Held = True
        
    def release(self):
        """
        Releses the queue, allowing new tasks to start
        """        
        self.Held = False
        self.start_tasks()
        
    @synchronized
    def is_empty(self):
        """
        Returns:
            bollean: True if no tasks are running and no tasks are waiting
        """
        return len(self.Queue) == 0
        
    isEmpty = is_empty
    
    @synchronized
    def waitUntilEmpty(self):
        """
        Blocks until the queue is empty (no tasks are running or waiting)
        """
        # wait until all tasks are done and the queue is empty
        if not self.is_empty():
            while not self.sleep(function=self.is_empty):
                pass
                
    join = waitUntilEmpty

    def drain(self):
        """
        Blocks until no more tasks are running
        """
        while self.nrunning() > 0:
            self.sleep(10)

    @synchronized
    def flush(self):
        """
        Discards all tasks. Running tasks will not be interrupted.
        """
        self.Queue.flush()

    @synchronized
    def cancel(self, task):
        """
        Cancel a queued task and remove it from the queue. The Promise associated with the Task will be cancelled.
        If the queue has a delegate, the taskCancelled delegate's method will be called.
        If the task is already running, it will not be interrupted.
        If the task is not found in the queue, ValueError exception will be raised.

        Args:
            task (Task): A Task subclass instance returned by the TaskQueue.append() or similar method
        
        Returns:
            Task: cancelled task
        """

        task.cancel()
        try:    self.Queue.remove(task)
        except ValueError:
            raise ValueError("Task not in the queue")
        self.call_delegate("taskCancelled", self, task)
        self.start_tasks()
        return task

    def __len__(self):
        """
        Returns total number of tasks in the queue, running and pending.
        """
        return len(self.Queue)

    def __contains__(self, task):
        """
        Returns true if the task is in the queue, running or waiting.
        """
        return task in self.Queue


class _Delegate(object):
    
    def task_failed(self, queue, task, exc_type, exc_value, tb):
        traceback.print_exception(exc_type, exc_value, tb, file=sys.stderr)

_GlobalTaskQueue = TaskQueue(delegate = _Delegate())

del _Delegate

def schedule_task(fcn, *params, **args):
    return _GlobalTaskQueue.append(fcn, *params, **args)



import time, random
from pythreader import TaskQueue, Task
from threading import Timer
import uuid

class MyTask(Task):

    def __init__(self):
        Task.__init__(self)
        self.Id = uuid.uuid4().hex[:5]

    def run(self):
        print (time.time(), ":", self.Id, "started")
        time.sleep(random.random()*0.5)
        print (self.Id, "ended")

q = TaskQueue(5, stagger=0.0001, capacity=30)

promises = [q.append(MyTask(), after=random.random()).promise for _ in range(100)]
for _ in range(100):
    time.sleep(random.random()/10)
    promises.append(q.append(MyTask(), after=random.random()).promise)

assert len(promises) == 200
for p in promises:
    p.wait()
    print("Queue counts:", q.counts())
print("All promises fulfilled")

q.waitUntilEmpty()
print("The queue is empty")



from .task_queue import TaskQueue, Task
from .core import Primitive, synchronized

class Gang(Primitive):

    def __init__(self, callable, params=None, n=1, concurrency=None, stagger=None, delegate=None):
        self.Queue = TaskQueue(concurrency, stagger=stagger, delegate=delegate or self)
        if params is None:
            params = [None] * n
        self.Promises = [self.Queue.add(callable, param).promise for param in params]

    def __iter__(self):
        return (promise.wait() for promise in self.Promises)

    def wait(self):
        return list(self.__iter__())



from .core import Primitive, synchronized, PyThread, gated, Timeout, Timer
from .dequeue import DEQueue
from .task_queue import TaskQueue, Task, schedule_task
from .Scheduler import Scheduler
from .Subprocess import ShellCommand
from .RWLock import RWLock
from .Version import Version
from .promise import Promise
from .processor import Processor
from .flag import Flag
from .gate import Gate
from .LogFile import LogFile, LogStream
from .producer import Producer
from .escrow import Escrow
from .gang import Gang

__version__ = Version
version_info = (tuple(int(x) for x in Version.split(".")) + (0,0,0))[:3]            # pad with 0s

__a_ll__ = [
    'Primitive',
    'PyThread',
    'TimerThread',
    'DEQueue',
    'gated',
    'synchronized',
    'Task',
    'TaskQueue',
    'Subprocess',
    'ShellCommand',
    'Version', '__version__', 'version_info',
    'Timeout',
    'Promise',
    'Scheduler',
    'Gate', 'LogFile', 'LogStream',
    'Escrow', 'Producer', 'Gang'
]



import time
from pythreader import Task, TaskQueue
from pythreader.core import printWaiting
from threading import Timer

import random

class T(Task):
	def __init__(self, id):
		Task.__init__(self)
		self.Id = id

	def run(self):
	    print ("task %d: t=%s" % (self.Id, time.ctime(time.time())))

tq = TaskQueue(1, capacity=10, stagger=1.0)

def printStats():
	with tq:
		printWaiting()


for i in range(100):
	t = T(i)
	print ("submitting task %d..." % (i,))
	tq << t
	print ("task %d submitted" % (i,))
	time.sleep(0.2+random.random())
	if i and i % 10 == 0:
		Timer(0.5, printStats).start()
		

tq.waitUntilEmpty()



from .core import Primitive

class Flag(Primitive):

    def __init__(self, value=False):
        Primitive.__init__(self)
        self._Value = value

    def set(self, value=True):
        self._Value = value
        self.wakeup()

    def get(self):
        return self._Value

    value = property(get, set)

    def sleep_until(self, value=None, predicate=None, timeout = None):
        if value is not None:
            predicate = lambda x, v=value: x == v
        if predicate is None:
            predicate = lambda x: bool(x)     # default: wait until the value evaluates to True
        while True:
            with self:
                if predicate(self._Value):
                    break
                else:
                    self.sleep(timeout)



from pythreader import Scheduler
import time, traceback
from datetime import datetime

s = Scheduler()

class NTimes(object):

    def __init__(self, n):
        self.N = n
    
    def __call__(self, message):
        t = datetime.now()
        print("%s: %s" % (t.strftime("%H:%M:%S.%f"), message))
        self.N -= 1
        if self.N <= 0:
            print("stopping", message)
            return "stop"
        
s.start()
time.sleep(1.0)
s.add(NTimes(10), 1.5, "green 1.5 sec")
print("+++added green")
time.sleep(1.0)
s.add(NTimes(7), 0.5, "blue 0.5 sec")
print("+++added blue")


time.sleep(20.0)
s.add(NTimes(10), 0.25, "red 0.25 sec", t0=0)
print("+++red green")

s.join()

        


import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), "r").read()

def get_version():
    g = {}
    exec(open(os.path.join("robotz", "Version.py"), "r").read(), g)
    return g["Version"]


setup(
    name = "robotz",
    version = get_version(),
    author = "Igor Mandrichenko",
    author_email = "igorvm@gmail.com",
    description = ("A set of useful tools built on top of standard Python threading module"),
    license = "BSD 3-clause",
    keywords = "threading, synchronization, multiprocessing, parallel computing",
    url = "https://github.com/imandr/robotz",
    packages=['robotz', 'tests'],
    long_description="A set of useful tools built on top of standard Python threading module", #read('README'),
    zip_safe = False,
    entry_points = { "console_scripts": [
                "director = director.director:main",
            ]
        },
    classifiers=[
        "Operating System :: POSIX",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3"
    ]
)



from .core import Primitive, synchronized
from threading import current_thread
import threading

def _thread_id():
    return current_thread().ident

class RWLock(Primitive):
    
    def __init__(self):
        
        """
        Creates read-write lock. Read-write lock has 2 ways to ackquire it: *exclusive* and *shared*.
        If the read-write lock is already acquired in a shared way, it can be acquired by any number of threads in shared way
        without blocking, but an attempt to lock it exclusively will block until the lock is completely released.
        If the lock is acquired in exclusive way, any other attempt to lock it, in exclusive or shared way will block.
        
        The lock is reentrant in the sense that if a thread acquires it in exclusive mode, it can ackquire it again in exclusive mode.
        If the lock is scquired multiple times by a thread, it has to be released the same number of times.
        
        If the lock is acquired in the shared mode by only one thread, and the same thread is trying to acquire it in exclusive mode,
        the thread will not block, and the thread will be granted exclusive lock. When the thread later releases the exclusive lock,
        the lock still remains locked by the thread in shared mode until explicitly released.
        """
        
        Primitive.__init__(self)
        self.Exclusive = None
        self.ExclusiveCount = 0
        self.ExclusiveQueue = []
        self.Shared = {}
        
    def __purge(self):
        threads = set([t.ident for t in threading.enumerate()])
        if not self.Exclusive in threads:
            self.Exclusive = None
            self.ExclusiveCount = None
        for t in list(self.Shared.keys()):
            if not t in threads:
                del self.Shared[t]
        self.ExclusiveQueue = [t for t in self.ExclusiveQueue if t in threads]
    
    def __acq_exclusive(self):
        tid = _thread_id()
        self.__purge()
        if tid == self.Exclusive:
            self.ExclusiveCount += 1
            return True
        if self.Exclusive is None and (len(self.Shared) == 0 or len(self.Shared) == 1 and tid in self.Shared):
            if self.ExclusiveQueue:
                if tid == self.ExclusiveQueue[0]:
                    self.ExclusiveQueue = self.ExclusiveQueue[1:]
                else:
                    self.ExclusiveQueue.append(tid)
                    return False
            self.Exclusive = tid
            self.ExclusiveCount = 1
            return True
        return False
        
    def __rel_exclusive(self):
        tid = _thread_id()
        assert self.Exclusive == tid and self.ExclusiveCount > 0
        self.ExclusiveCount -= 1
        if self.ExclusiveCount <= 0:
            self.ExclusiveCount = 0
            self.Exclusive= None
        self.wakeup(all=True)
        
    def __acq_shared(self):
        tid = _thread_id()
        self.__purge()
        if not self.Exclusive in (None, tid):
            return False
        if not tid in self.Shared:
            self.Shared[tid] = 0
        self.Shared[tid] = self.Shared[tid] + 1
        return True
            
    def __rel_shared(self):
        tid = _thread_id()
        n = self.Shared.get(tid)
        assert n is not None and n > 0
        n -= 1
        if n == 0:
            del self.Shared[tid]
        else:
            self.Shared[tid] = n
        self.wakeup(all=True)
        
    @synchronized
    def acquireExclusive(self):
        while not self.__acq_exclusive():
            self.sleep()
            
    @synchronized
    def acquireShared(self):
        """
        Ackquires the lock in shared way. Will block if the lock is acquired by some other thread in exclusive mode.
        If the lock is exclusively locked by the same thread, the lock will also be locked in shared way and the method will return immediately.
        """
        while not self.__acq_shared():
            self.sleep()
            
    @synchronized
    def releaseExclusive(self):
        """
        Removes the exclusive lock if the calling thread holds it. If the lock was exclusively acquired multiple times by the thread,
        it will remain exclusively locked by the thread until it is exclusively-released the same number of times. 
        """
        self.__rel_exclusive()
        
    @synchronized
    def releaseShared(self):
        """
        Removes the shared lock if the calling thread holds it. If the lock was exclusively acquired multiple times by the thread,
        it will remain locked in shared mode by the thread until it is exclusively-released the same number of times. 
        """
        self.__rel_shared()

    @property
    def exclusive(self):
        """
        Create *context* object, which will acquire the lock in exclusive mode on entry and release it on exit
        For example:
        
            lock = RWLock()
        
            with lock.exclusive:
                # now the lock is exclusively locked
                ...
            # exit from the context removes the exclusive lock
        """
        
        class ExclusiveContext(object):
            def __init__(self, rwlock):
                self.RWLock = rwlock
            
            def __enter__(self):
                return self.RWLock.acquireExclusive()
            
            def __exit__(self, t, v, tb):
                return self.RWLock.releaseExclusive()
        return ExclusiveContext(self)
    
    @property
    def shared(self):
        """
        Create *context* object, which will acquire the lock in shared mode on entry and release it on exit
        For example:
        
            lock = RWLock()
        
            with lock.shared:
                # now the lock is locked in shared way
                ...
            # exit from the context removes the shared lock
        """
        class SharedContext(object):
            def __init__(self, rwlock):
                self.RWLock = rwlock
            
            def __enter__(self):
                return self.RWLock.acquireShared()
            
            def __exit__(self, t, v, tb):
                return self.RWLock.releaseShared()
            
        return SharedContext(self)
        
    @synchronized
    def owners(self):
        return self.Exclusive, list(self.Shared.keys())
        



from .core import Primitive, synchronized

class Escrow(Primitive):

    def __init__(self):
        Primitive.__init__(self)
        self.WaitingGet = False
        self.Value = None
        self.GetLock = Primitive()
        self.PutLock = Primitive()

    def get(self, timeout = None):
        with self.GetLock:
            try:
                self.WaitingGet = True
                self.wakeup()                   # notify the thread, which it waiting to put
                while self.Value is None:
                    self.sleep(timeout)         # will raise exception on timeout
                value = self.Value
                self.Value = None
                self.wakeup()                   # tell the putter that the package was received
                return value
            finally:
                self.WaitingGet = False

    def put(self, value, timeout = None):
        assert value is not None
        with self.PutLock:
            try:
                while not self.WaitingGet:      # wait for the receiver
                    self.sleep(timeout)
                self.Value = value
                self.wakeup()
                while self.Value is not None:   # wait for pickup
                    self.sleep()
            finally:
                self.Value = None               # time-out or something else

class EscrowBarrier(Primitive):
    
    def __init__(self, threshold=2, initial_value=None, **args):
        Primitive.__init__(self, **args)
        self.Value = self.InitialValue = initial_value
        self.Threshold = threshold
        self.NWaiting = 0
        self.Done = False
    
    @synchronized
    def arrive(self, value=None, increment=False, timeout=None):
        if not self.Done:
            if increment:
                self.Value += value
            else:
                self.Value = value
            self.NWaiting += 1
            while self.NWaiting < self.Threshold:
                self.sleep(timeout)
            self.Done = True
            self.NWaiting = 0
        return self.Value

    @synchronized
    def wait(self, timeout=None):
        while not self.Done:
            self.sleep(timeout)
        return self.Value



from .core import PyThread, synchronized, Primitive, Timeout
from .task_queue import Task, TaskQueue
from .promise import Promise
import time, uuid, traceback, random
import sys

class Job(object):
    
    def __init__(self, scheduler, id, t, interval, jitter, fcn, count, params, args):
        self.F = fcn
        self.Params = params or ()
        self.Args = args or {}
        self.ID = id
        self.Interval = interval
        self.Jitter = jitter or 0.0
        self.NextT = t
        self.Scheduler = scheduler
        self.Count = count
        self.Promise = None
        
    def __str__(self):
        return f"Job({self.ID})"
        
    __repr__ = __str__
        
    def run(self):
        promise = self.Promise
        self.Promise = None
        self.Scheduler = None           # to break circular dependencies
        start = time.time()
        exc_info = None
        try:    
            #print(f"Job({self.ID}): running:", self.F, self.Params, self.Args)
            next_t = self.F(*self.Params, **self.Args)
            promise.complete()
        except Exception as e:
            #print(f"Job({self.ID}):", e)
            exc_info = sys.exc_info()
            #print(f"Job({self.ID}): exception:", traceback.format_exc())
            promise.exception(*exc_info)
            next_t = None
        else:
            #print(f"Job({self.ID}): done")
            pass
        if self.Count is not None:
            self.Count -= 1
            if self.Count <= 0:
                return None, exc_info
        if next_t == "stop":
            next_t = None
        elif next_t is None and self.Interval is not None:
            next_t = start + self.Interval + random.random() * self.Jitter
        if next_t is not None and next_t < 3.0e7:
            # if next_t is < 1980, it's relative time
            next_t = next_t + start + random.random() * self.Jitter
        return next_t, exc_info
        
    def cancel(self):
        """Cancels the job. If the job is currently running, it will not be stopped, but if it's a repeating job, calling this
        method will prevent the job from running again.
        """
        try: 
            self.Scheduler.remove(self)
        except: pass
        self.Scheduler = None

class JobThread(PyThread):
    
    def __init__(self, scheduler, job):
        PyThread.__init__(self, name=f"Scheduler({scheduler})/job{job.ID}", daemon=True)
        self.Scheduler = scheduler
        self.Job = job
        
    def run(self):
        scheduler = self.Scheduler
        job = self.Job
        self.Job = self.Scheduler = None
        next_t, exc_info = job.run()
        if exc_info:
            scheduler.job_failed(job, next_t, *exc_info)
        else:
            scheduler.job_ended(job, next_t)

class Scheduler(PyThread):
    def __init__(self, max_concurrent = 100, stop_when_empty = False, delegate=None, daemon=True, name=None, start=True, **args):
        """
        Args:
            max_concurrent (int): maximum number of concurrent jobs to run. Default: 100
            stop_when_empty (bool): stops the Scheduler thread when all the jobs complete. Default: False
            delegate (object): an object to notify when a job ends or fails
            daemon (bool): whether the Scheduler thread will run as daemon. Default: False
            name (str): name of the Scheduler
            start (bool): whether to start the Scheduler immediately on initialization. If False, the Scheduler needs to be started
                by calling ``start()`` method. Default: True
        """
        PyThread.__init__(self, daemon=daemon, name=name)
        self.Timeline = []      # [job, ...]
        self.Delegate = delegate
        self.StopWhenEmpty = stop_when_empty
        self.Stop = False
        if start:
            self.start()

    # delegate interface used to wait until the task queue is empty
    def job_ended(self, job, next_t):
        #print("job_ended:", job.ID)
        if self.Delegate is not None and hasattr(self.Delegate, "jobEnded"):
            try:    self.Delegate.jobEnded(self, task.JobID)
            except: pass
        if next_t is not None:
            self.add_job(job, next_t)
        self.wakeup()

    def job_failed(self, job, next_t, exc_type, exc_value, tb):
        #print("job_failed:", job.ID)
        if self.Delegate is not None and hasattr(self.Delegate, "jobFailed"):
            try:    self.Delegate.jobFailed(self, job.ID, exc_type, exc_value, tb)
            except: pass
        if next_t is not None:
            self.add_job(job, next_t)
        self.wakeup()
        
    @synchronized
    def jobs(self):
        return self.Timeline[:]

    def stop(self):
        """Stops the Scheduler thread"""
        self.Stop = True
        self.wakeup()
        
    @synchronized
    def add_job(self, job, t):
        job.NextT = t
        self.Timeline.append(job)
        job.Promise = promise = Promise(job).oncancel(job.cancel)
        self.wakeup()
        return promise
        
    @synchronized        
    def add(self, fcn, *params, interval=None, t=None, t0=None, id=None, jitter=0.0, param=None, count=None, **args):
        """Adds a new job to the schedule.
        
        Args:
            fcn (callable): will be called when the job starts
            params: positional arguments to pass to ``fcn``
            args: keyword arguments to to pass to ``fcn``
            interval (numeric): interval to repeat the job. Default: None (do not repeat)
            count (int): if ``interval`` is specified, specifies how many times to repeat the job. Default: unlimited
            t (numeric): pecifies the time when to fire the Timer first time. ``t`` can be either absolute timestamp - time in seconds since the
                Epoch or relative to current time. If ``t`` is less than 3e8 (~10 years), then it is interpreted as relative to current time.
                Default: current time plus ``interval``
            id (str): id to assign to the job. If None, the id will be generated automatically
            t0 - deprecated, alias to t, retained for backward compatibility
            param: deprecated, single positional parameter to pass to ``fcn``. Use ``params`` instead        
        
        Returns:
            Job: job object
        """
        #
        # t0 - first time to run the task. Default:
        #   now + interval or now if interval is None
        # interval - interval to repeat the task. Default: do not repeat
        #
        # fcn:
        #   next_t = fcn()
        #   next_t = fcn(param)
        #   next_t = fcn(**args)
        # 
        #   next_t:
        #       "stop" - remove task
        #       int or float - next time to run
        #       None - run at now+interval next time
        #
        if t is None: t = t0            # alias, t0 argument is deprecated
        if param is not None:       # for backward compatibility
            params = (param,)
        if id is None:
            id = uuid.uuid4().hex[:8]
        if t is None:
            t = time.time() + (interval or 0.0) + random.random()*jitter
        elif t < 10*365*24*3600:           # ~ Jan 1 1980
            t = time.time() + t
        job = Job(self, id, t, interval, jitter, fcn, count, params, args)
        return self.add_job(job, t)
        
    @synchronized
    def remove(self, job_or_id):
        """Removes the job from the schedule. Will not stop the job if it is already running.
        
        Args:
            job_or_id: Either the Job object returned by the ``add`` method, or job id (str)
        """
        job_id = job_or_id if isinstance(job_or_id, str) else job_or_id.ID
        self.Timeline = [j for j in self.Timeline if j.ID != job_id]

    @synchronized
    def run_jobs(self):
        keep_jobs = []
        next_run = None
        for job in self.Timeline:
            if job.NextT <= time.time():
                t = JobThread(self, job)
                t.start()
                self.wakeup()
            else:
                next_run = min(next_run or job.NextT, job.NextT)
                keep_jobs.append(job)
        self.Timeline = keep_jobs
        return next_run

    @synchronized
    def is_empty(self):
        """Returns True if there are no pending jobs on the schedule. Note that if a repeating job is currently running, it will not
        be on the schedule until it completes or fails.
        """
        return not self.Timeline
        
    
    @synchronized
    def _next_run(self):
        # returns run time of first job to run
        if self.Timeline:
            return min(j.NextT for j in self.Timeline)
        else:
            return None

    @synchronized
    def _job_is_ready(self):
        return any(j.NextT <= time.time() for j in self.Timeline)

    def run(self):
        while not self.Stop and not (self.is_empty() and self.StopWhenEmpty):
            with self:
                delta = 100
                if self.Timeline:
                    next_t = self.run_jobs()
                    if next_t is not None:
                        delta = next_t - time.time()
                self.sleep(delta)

    @synchronized        
    def wait_until_empty(self):
        """Wait until the schedule is empty. Note that if a repeating job is currently running, it will not
        be on the schedule until it completes or fails.
        """
        while not self.is_empty():
            self.sleep(10)
    
    join = wait_until_empty

_GLobalSchedulerLock = Primitive()
_GlobalScheduler = None

def global_scheduler(name="GlobalScheduler", **args):
    global _GlobalScheduler
    if _GlobalScheduler is None:
        with _GLobalSchedulerLock:
            if _GlobalScheduler is None:
                _GlobalScheduler = Scheduler(name=name, **args)
    return _GlobalScheduler

def schedule_job(fcn, *params, **args):
    return global_scheduler().add(fcn, *params, **args)

def unschedule_job(job_or_id):
    return global_scheduler().remove(job_or_id)

schedule_task = schedule_job        # aliases, deprecating term "job"
unschedule_task = unschedule_job

if __name__ == "__main__":
    from datetime import datetime
    
    s = Scheduler()
    
    class NTimes(object):

        def __init__(self, n):
            self.LastRun = None
            self.N = n

        def __call__(self, message):
            t = datetime.now()
            delta = None if self.LastRun is None else t - self.LastRun
            print("%s: %s delta:%s" % (t.strftime("%H:%M:%S.%f"), message, delta))
            self.LastRun = t
            self.N -= 1
            time.sleep(0.2)
            if self.N <= 0:
                print("stopping", message)
                return "stop"

    #time.sleep(1.0)
    s.add(NTimes(10), "green 1.5 sec", interval=1.5)
    s.add(NTimes(7), "blue 0.5 sec", interval=0.5)
    s.start()

    print("waiting for the scheduler to finish...")
    #s.join()
    s.wait_until_empty()
    s.stop()
    s.join()
    print("stopped")



from pythreader import Processor
import time, random

class MyProcessor(Processor):
    
    def process(self, x):
        time.sleep(random.random()*0.5)
        return x**2
        
processor = MyProcessor(5)

promises = [processor.put(i) for i in range(100)]
for p in promises:
    print(p.Data, "->", p.wait())